/*
   csg.cc -- this implements CSG. 

   (c) 1993, 1996 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

 */


#include "object.hh"
#include "nonprim.hh"
#include "proto.hh"
#include "extern.hh"
#include "ibuf.hh"


/*
   intersect a CSG union.

   Actually I could use the same code to do CSG intersections, since

   intersection { { A inverse } { B inverse } inverse }

   is equivalent to

   union { A B }

   My algorithm relies on the Jordan curve theorem (Gee, that sounds great...
   ) Well kindof... Jordan curves have to divide space into a *bounded*
   and unbouded set.;

   Anyway, this works because each crossed boundary means we're going
   from inside->outside or outside->inside of the CSG member, thus
   making the number of shapes we're in very easy to keep track of.

    Disadvantage: the shapes to be used should be "closed", so plane -> ok,
   polygon -> nope
 */

bool    union_data::all_intersections(ibuf * globalq, ray * r,
                                              int flags, bool &isinside) const
{
    stat->test++;

    bool chkall = flags & CHKALL;
    int    encl_shapes = 0;    /* shapes we're inside of. */
    int    total = 0;
    ibuf   CSGq;	/* set up a new queue for results of this CSG union. */
    int intersect_flags = flags | CHKINSIDE | CHKALL;

    /* count the shapes and intersect them. */
    for (object_iter p(contents); p.ok(); p.next())
    {
	bool local_inside;            /* found an intersection? */
	total++;
	bool gotcha = p->all_intersections(&CSGq, r, intersect_flags, local_inside);

        /* try to get away cheaply */
        if (!gotcha && local_inside)
        {
            isinside = !(daddy->flags & O_INVERTED);
            return false;
        }
        if (local_inside)
            encl_shapes++;
    }

    /* encl_shapes now holds the number of shapes we're in
     *
     * during the following loop, it holds the number of shapes we're in, after we've passed intersection q
     *
     * During the loop, local_inside is wether we are inside or outside the CSG. */

    /* ray origin inside CSG? */
    bool local_inside = isinside = (encl_shapes==0) ^ bool(daddy->flags & O_INVERTED);
    bool gotcha = false;

    for (ibuf_iter q(CSGq); q && (chkall || !gotcha); q.next())
    {
        if (q->entering)
        {                       /* going into a shape */
            ++encl_shapes;

	    // we can't be in more shapes than there are shapes to be in (huh :-)
            assert(encl_shapes <= total);

	    // from outside CSG -> inside CSG, this is an intersection 
            if (encl_shapes == 1)
            {
                local_inside = !local_inside;
                q->entering = local_inside;
                globalq->add_to(flags, *(dqueue *)q);
                gotcha = true;
            }
        } else {
            /* exiting a shape */
            --encl_shapes;
            assert(encl_shapes >= 0);

            if (encl_shapes == 0)
            {                   /* just left the CSG structure */
                local_inside = !local_inside;
                q->entering = local_inside;
                globalq->add_to(flags, *(dqueue *)q);
                gotcha = true;
            }
        }
    }

    if (gotcha)
        stat->hit++;

    return gotcha;
}

bool    inter_data::all_intersections(ibuf * globalq, ray * r,
                                              int flags, bool &isinside) const
{
    stat->test++;
    bool chkall = flags & CHKALL;

    /* count the shapes, and intersect them. */
    int encl_shapes = 0;
    int total = 0;
    
    /* new queue for results of this CSG union. */
    ibuf   CSGq;

    int intersect_flags = flags | CHKINSIDE | CHKALL;
    
    /* walk all the  CSG elements, and intersect each one */
    for (object_iter p(contents); p.ok(); p.next())
    {
        total++;
	bool local_inside;
	bool        gotcha = p->all_intersections(&CSGq, r, intersect_flags, local_inside);

        /* we're outside, we don't hit this one. */
        if (!gotcha && !local_inside)
        {
            isinside = bool(daddy->flags & O_INVERTED);
            return false;
        }
        if (local_inside)
            encl_shapes++;
    }

    /* encl_shapes now holds the number of shapes we're in
     *
     * during the following loop, it holds the number of shapes we're in, after we've passed intersection "q"
     *
     * During the loop, isinside is wether we are inside or outside the CSG shape.
     */

    /* ray origin inside CSG? */
    bool local_inside = (encl_shapes == total) ^ bool(daddy->flags & O_INVERTED);
    isinside = local_inside;
    
    bool gotcha = false;


    for (ibuf_iter q(CSGq); q && (chkall || !gotcha); q.next())
    {
        if (q->entering)
        {	/* going into a shape */
            ++encl_shapes;

	    /* we can't be in more shapes than there are shapes to be in (huh :-) */
            assert(encl_shapes <= total);

            if (encl_shapes == total)
            {                   /* from outside CSG -> inside CSG, this is an intersection */
                local_inside = !local_inside;
                q->entering = local_inside;
                globalq->add_to(flags, *(dqueue *)q);
                gotcha = true;
            }
	} else {                       /* leaving shape */
            if (encl_shapes == total)
            {                   /* just left the CSG structure */
                local_inside = !local_inside;
                q->entering = local_inside;
                globalq->add_to(flags, *(dqueue *)q);
                gotcha = true;
            }

            /* exiting a shape */
            --encl_shapes;
            assert(encl_shapes >= 0);
        }
    }
    if (gotcha)
        stat->hit++;

    return gotcha;
}

/* inside?  true if we're at least inside one of the components */
bool
union_data::inside(vector x) const
{
    for (object_iter p(contents); p.ok(); p.next())
        if (p->inside(x))
            return true;
    return false;
}


/* inside of CSGinter if we're inside all parts of CSG */
bool
inter_data::inside(vector x) const
{
    for (object_iter p(contents); p.ok(); p.next())
        if (!p->inside(x))
            return false;
    return true;
}

void
inter_data::print() const
{
#ifdef DEBUG
    contents.print();
#endif          /* DEBUG */
}
void
union_data::print() const
{
#ifdef DEBUG
    contents.print();
#endif          /* DEBUG */
}
/*
   QUESTION: from what file were the routines below copied? (u gaat door
   voor de koelkast!)
 */

/* add an object to the CSG c */
void
CSG_data::add_child(object * o)
{
    o->daddy = daddy;
    contents.add_to(o);
}

void
CSG_data::translate(const vector t)
{
    contents.translate(t);
}

/* hooooeeewaahhh, Geez, why do I write these comments? */
void
CSG_data::rotate(const matrix & rotmat)
{
    contents.rotate(rotmat);
}

/*
   I guess .. because somebody will read them, some day (Hello, is there
   somebody out there?

   and another destructor (en wat hebben ze gewonnen, Pierre, als ze goed
   raden waar dit voor is ?
 */

/* 3X raden (Yes, you folks out there, I'm now commenting in Dutch! */
void
CSG_data::scale(const vector s)
{
    contents.scale(s);
}

/* feeble attempt to speed up csg intersection */
void
inter_data::bound(matrix *, bounding_box *bbox)
{
    contents.bound();

    if (!(daddy->flags & O_INFINITE))
	/* Try to bound the intersection.  Although it won't be optimal, it's better than nothing. */
	for (object_iter op(contents); op.ok(); op.next())
	    /* if has a finite inside, any intersection will be bounded */
	    if (op->flags & O_FINITE_INSIDE)
		bbox->intersect(*op->bbox);
}

void
inter_data::precompute()
{
    stat->howmuch++;
    contents.precompute(daddy);
    
    int flags = daddy->flags;
    flags |= O_INFINITE;
    flags &= ~O_FINITE_INSIDE;

    /* Try to bound the intersection.  Although it won't be optimal, it's better than nothing. */
    for (object_iter op(contents); op.ok(); op.next())
    {
        /* if has a finite inside, any intersection will be bounded */
        if (op->flags & O_FINITE_INSIDE) {
            flags &= ~O_INFINITE;
            flags |= O_FINITE_INSIDE;
        }
    }
    nonprims++;
    daddy->flags = flags;
}

/* feeble attempt to speed up CSG union */
void
union_data::bound(matrix *, bounding_box *bbox)
{
    contents.bound();
    bbox->set_empty();

    if (!(daddy->flags & O_INFINITE))
	for (object_iter op(contents); op.ok(); op.next())

	    /* if any has a infinite inside, the union will be infinite */
	    if (!(op->flags & O_FINITE_INSIDE)) {

		    bbox->set_full();
		    break;
		} else
		    bbox->add(*op->bbox);

}

void
union_data::precompute()
{
    stat->howmuch++;
    contents.precompute(daddy);
    int flags = daddy->flags;
    flags &= ~O_INFINITE;
    flags |= O_FINITE_INSIDE;
    for (object_iter op(contents); op.ok(); op.next())
    {
        /* if any has a infinite inside, the union will be infinite */
        if (!(op->flags & O_FINITE_INSIDE))
        {
            flags |= O_INFINITE;
            flags &= ~O_FINITE_INSIDE;
	    break;
        }
    }
    daddy->flags =  flags;
    nonprims++;
}

shape_stats *union_data::stat = shape_stats::regist("CSG union");
shape_stats *inter_data::stat = shape_stats::regist("CSG inter");

PUBLIC object *
        get_new_inter_object()
{
    object *o = new object;
    inter_data *C = new inter_data;
    o->data = C;
    o->data->type = CSGINTER;
    C->setdaddy(o);
    o->flags = O_INFINITE | O_CLOSED;

    return o;
}

PUBLIC object *
        get_new_union_object()
{
    object *o = new object;
    union_data *C = new union_data;
    o->data = C;
    o->flags = O_INFINITE | O_CLOSED;
    o->data->type = CSGUNION;
    C->setdaddy(o);
    return o;
}
