/*
   disc.c -- handle disc type objects

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */


#include "object.hh"
#include "shapes.hh"
#include "proto.hh"
#include "extern.hh"
#include "matrix.hh"
#include "ibuf.hh"


/* initialize a disc_data */
shape_stats *disc_data::stat = shape_stats::regist("disc");
disc_data::disc_data(vector c, vector in, flt R, flt r)
{
    type = DISC;
    n =in;
    center=c;
    majr = R;
    minr = r;
}

/* scale it. */
void
disc_data::scale(const vector s)
{
    n.x = n.x / s.x;
    n.y = n.y / s.y;
    n.z = n.z / s.z;

    flt factor = abs(s.x);   
    majr *= factor;
    minr *= factor;
    center.scale(s);
}

/* rotation */
void
disc_data::rotate(const matrix & rotmat)
{
    n.rotate(rotmat);
    center.rotate(rotmat);
}

/* and translation */
void
disc_data::translate(const vector t)
{
    center += t;
}


void
disc_data::print() const
{
#ifdef DEBUG
    n.print("normal");
    center.print("center");
    dprintf("majr%f r %f\n", majr, minr);
#endif          /* DEBUG */
}
/*
 * intersect with the this part, return true if intersection
 * found.
 */
bool    disc_data::all_intersections(ibuf * q, ray * r, int flags, bool &) const
{    
    stat->test++;

    /* do a plane intersection */
    flt    dot = r->dir * n;

    if (ISZERO(dot))
        return false;

    flt t = (mov - r->pos * n) / dot;
    if (t < tolerance || t > r->maxt)
        return false;

    /* find distance to center */
    vector      x = r->point(t) - center;
    flt         l = x * x;

    /* inside? */
    if (l < Rsq && l > rsq)
    {
        stat->hit++;
	dqueue  q_ent;
        q_ent.obj = daddy;
        q_ent.t = t;
        q->add_to(flags, q_ent);
        return true;
    } else
        return false;
}
/*
 * returns the normal to the disc
 */
vector  disc_data::normal(vector) const
{
    return n;
}
/*
 * bound a disc. It's the same routine as the torus, except, the the minor
 * radius is set to zero.
 */

void
disc_data::bound(matrix *trans, bounding_box *bbox)
{
    flt        l = (n).length();
    flt             theta = acos(n.y / l);
    flt             phi = safe_arctangent(n.z, n.x);

    vector rot(0, 0, -theta);
    matrix  ctm;
    ctm.rotate(rot);
    
    rot.set(0, -phi, 0);
    matrix tempmatrix;
    tempmatrix.rotate(rot);
    ctm = tempmatrix * ctm;
    tempmatrix.translate(center);
    tempmatrix = tempmatrix * ctm;

    if (trans)    
        tempmatrix = (*trans) * (tempmatrix);
    
    bound_torus(bbox, majr, 0, tempmatrix);
}

void
disc_data::precompute()
{
    stat->howmuch++;
   prims++;
 
    mov = n * center;
    if (minr <= 0)
        rsq = -1;
    else
        rsq = sqr(minr);
    Rsq = sqr(majr);
}


/* alloc/init */
PUBLIC object *
        get_new_disc_object(vector c, vector in, flt R, flt r=-1)
{
    object *o = new object;
    o->data = new disc_data(c,in,R,r);
    o->data->setdaddy(o);
    o->flags |=  O_ONLY_UNIFORM_SCALE ;
    return o;
}
/* eof */
