/*
   grids.c -- handle grids and gridtrees

   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */

#include	"ray.hh"
#include	"proto.hh"
#include   	"shapes.hh"
#include 	"extern.hh"
#include	"light.hh"
#include	"grid.hh"
#include	"ibuf.hh"

// complicated stuff, trying to avoid aliasing AND save time
Grid_tree::~Grid_tree()
{
    delete[] reflect;
    delete[] refract;
    delete[] indices;
    delete[] shadowindices;
}

/*
  alloc a new grid_tree entry, and construct a permutation on the global
  grid's indices
 */

Grid_tree::Grid_tree(Grid * r, int depth)
{
    root = r;
    reflect = refract = 0;
    indices = new int[r->grid_len];
    shadowindices = new int[r->grid_len];
    recursion = depth;
    last = shadowlast = 0;

    bool   *tmp_indices = new bool[r->grid_len];

    /* the grid */
    for (int i = 0; i < r->grid_len; i++)
        tmp_indices[i] = false;

    for (int  i = 0; i < r->grid_len; i++)
    {
	int j;
        do
        {
            j = rand() % r->grid_len;   /* guess an index */
        } while (tmp_indices[j]);       /* have we used it earlier? */
        indices[i] = j;
        tmp_indices[j] = true;
    }

    /* shadowray grid */
    for (int i = 0; i < r->grid_len; i++)
        tmp_indices[i] = false;

    for (int i = 0; i < r->grid_len; i++)
    {
	int j;
        do
        {
            j = rand() % r->grid_len;   /* guess an index */
        } while (tmp_indices[j]);       /* have we used it earlier? */
        shadowindices[i] = j;
        tmp_indices[j] = true;
    }

    delete  tmp_indices;
};

/* sample (u,v) from the grid entry g for a ray of type raytype

   PARAMS OUT
   u, v: pseudo Poisson on [-1,1]
   f: is a filter function int _[-1,1]^2 f dsigma  = 1

   PARAMS IN
   type of ray.

   EFFECTS
   will allocate appropriate child of this node.
 */
void
Grid_tree::sample(flt &u, flt & v, flt & f, int rtype)
{
    Grid_tree *child;
    int     index;
    switch (rtype)
    {
    case R_SHADOW:
        index = shadowindices[shadowlast];
        shadowlast = (shadowlast + 1) % root->grid_len;
        break;
    case R_REFLECT:
        if (!reflect)
            reflect = new Grid_tree(root, recursion + 1);
        child = reflect;

        index = child->indices[child->last];
        child->last = (child->last + 1) % root->grid_len;
        break;

    case R_REFRACT:
        if (!refract)
            refract = new Grid_tree(root, recursion + 1);
        child = refract;

        index = child->indices[child->last];
        child->last = (child->last + 1) % root->grid_len;
        break;

    case R_EYE:
        /* access global tree directly */
        index = indices[last];
        last = (last + 1) % root->grid_len;
        break;
    default:
        assert(false);
    }

    flt    *dp = root->the_grid + 2 * index;

    u = *dp++;
    v = *dp++;

    u += root->jitter_width_x * rand1();
    v += root->jitter_width_y * rand1();

    assert(abs(u) <= 1);
    assert(abs(v) <= 1);

    f = root->filter[index];
}

void
Grid::sample(flt&  u, flt &v, flt & f, int rtype) 
{
    tree->sample(u, v, f, rtype);
}

/* print out info from tree node g, and recurse */
void
Grid_tree::print() const
{
#ifdef DEBUG
    int     i;
    dprintf("\nindices: ");
    for (i = 0; i < root->grid_len; i++)
    {
        dprintf("%d ", indices[i]);
    }
    dprintf(" shadow indices: ");
    for (i = 0; i < root->grid_len; i++)
    {
        dprintf("%d ", shadowindices[i]);
    }
    dprintf("\n");

    if (refract)
    {
        dprintf("refract grid {");
        refract->print();
        dprintf("}\n");
    }
    if (reflect)
    {
        dprintf("reflect grid {");
        reflect->print();
        dprintf("}\n");
    }
#endif
}

/* print all there is to know about grids */
void
Grid::print() const
{
#ifdef DEBUG
    dprintf("%d entries, maxjitter (%f,%f) \n2d grid {\n", grid_len, jitter_width_x, jitter_width_y);
    flt *    dp = the_grid;
    for (    int     i = 0; i < grid_len; i++)
    {
        flt u = *dp++;
        flt v = *dp++;
        dprintf("(%f, %f)", u, v);
    }
    dprintf("}\nfilter {\n");
    dprintf("     type %d\n", filter_type);

    dp = filter;
    for (int i = 0; i < grid_len; i++)
    {
        dprintf("%f ", *dp++);
    }
    dprintf("}\n");

    tree->print();
#endif
}


Grid::Grid()
{
    jitter_width_x = jitter_width_y = 1.0;
    filter = 0;
    the_grid = 0;
    filter_type = NO_FILTER;
}
Grid::~Grid()
{
    delete[] filter;
    delete[] the_grid;
}
/* init a  grid */
void
Grid::precompute()
{
    /* try to make a square grid */
    int gridx = int (floor(sqrt(the_scene.tries)));
    int gridy = the_scene.tries / gridx;
    grid_len = gridx * gridy;

    flt    *gp = the_grid = new flt[2 * grid_len];
    flt    *fp = filter = new flt[grid_len];

    /* centers of a {0,1, .. ,  n-1} interval */
    flt    xcenter = ((flt) gridx - 1) / ((flt) 2 * gridx);
    flt ycenter = ((flt) gridy - 1) / ((flt) 2 * gridy);


    if (the_scene.tries == 1)
    {     // jittering looks really awful at 1 sample/pixel.
        jitter_width_x = 0.0;
        jitter_width_y = 0.0;
    } else
    {
        jitter_width_x *= 1 / ((flt) gridx);
        jitter_width_y *= 1 / ((flt) gridy);
    }

    /* make grid on [-1,1] */
    for (int x = 0; x < gridx; x++)
        for (int y = 0; y < gridy; y++)
        {
            flt xentry = 2 * ((flt) x / (flt) gridx - xcenter);
	    flt yentry = 2 * ((flt) y / (flt) gridy - ycenter);

            switch (filter_type)
            {
            case GAUSS_FILTER:
                *fp++ = exp(-(sqr(xentry) + sqr(yentry)));
                break;
            case NO_FILTER:
            default:
                *fp++ = 1;
                break;
            }

            *gp++ = xentry;
            *gp++ = yentry;
        }

    /* normalize */
    filterconst = 0;
    for (int i = 0; i < grid_len; i++)
        filterconst += filter[i];
    filterconst = grid_len / filterconst;
    for (int i = 0; i < grid_len; i++)
        filter[i] *= filterconst;
    tree = new Grid_tree(this, 0);
}


