/* -*-c++-*-
   intersection buffer macros
 */
/**************************************************************************/
/* the intersection structure */

#ifndef IBUF_H
#define IBUF_H

#include "vector.hh"

enum
{
    CHKALL = 0x01,              /* generate all intersections */
    CHKINSIDE = 0x02,           /* determine wether */
    USEMAIL = 0x04,             /* update mailboxes? */
    NO_CHKBOUNDS = 0x08,        /* check the bounding shapes ? */
    NO_TRANSUPDATE = 0x10       /* don't store transparent objects as minimal intersections */
};
/* ray types */
enum raytype
{
    R_SHADOW = 0x01,
    R_REFLECT = 0x02,
    R_REFRACT = 0x04,
    R_EYE = 0x08
};
/* object flags types */
enum
{
    O_NO_SHADOW = R_SHADOW,     /* invisible for shadowrays */
    O_NO_REFLECT = R_REFLECT,   /* invisible for reflectionrays */
    O_NO_REFRACT = R_REFRACT,   /* invisible for refraction */
    O_NO_EYE = R_EYE,           /* invisible for eyerays */
    O_TRANSPARENT = 0x10,       /* with tranparent texture (needs all intersections/inside check) */
    O_INFINITE = 0x20,          /* objects boundary is infinite (eg plane) */
    O_INVIS = 0x40,             /* invisible (lightsource) */
    O_CLOSED = 0x80,            /* closed surface */
    O_FINITE_INSIDE = 0x100,    /* object's inside is finite, eg sphere, NOT inverted sphere */
    O_FINITE_OUTSIDE = 0x200,   /* object's inside is infinite, eg inverted sphere */
    O_INVERTED = 0x400,         /* for CSG purposes. */
    O_ONLY_UNIFORM_SCALE = 0x800,
    O_STURM = 0x1000		// use sturm solver
};


struct raystatstruct
{
    long    shadow,             /* shadow test */
            reflected,          /* reflected rays */
            refracted,          /* refracted rays */
            eye;                /* eye rays */
    shape_stats *rays;
};


// linestats
struct linestatstruct
{                               // idem, for each line.
    long int rays,
            rayintersections,
            shadow,
            reflected,
            refracted,
            eye;
    int     maxdepth;
};


/* the ray structure */
class ray
{
public:
    long    rayid;              /* number of ray */
    flt     now;                // the time it was spawned.
    object *sender;
    vector  pos;                /* origin */
    vector  dir;                /* direction */
    raytype type;               /* what kind, see above */

    flt     maxt,               /* biggest value to look for (if it is more, than don't check intersection details */
            mint;               /* smallest t values to look for. (if intersection < mint, then stop searching other intersections.
                                 * Useful for shadow rays */
    static long lastrayid;
    static linestatstruct line;
    static raystatstruct stats;
    void operator=(ray const & src);
    
            ray(raytype newraytype, flt time, object * o = 0);
    vector  point(flt t)
    {                           // t * dir + pos
        vector  x = dir;
                x *= t;
                x += pos;
                return x;
    }
    // stats
    static void  reset_stats();
    void print() const ;
};


/* Depth queue */
struct dqueue
{
    object *obj,                /* the object which was hit. */
           *obj2;               /* needed for extrusion */
    flt     t;                  /* where in the ray */
    bool    entering;           /* entering or exiting? */
    dqueue() {}
    dqueue(flt it, object *io, bool ient =false,object * io2 = 0) {
	obj = io;
	entering = ient;
	obj2 =io2;
	t = it;	
    }
};

/* structure for passing intersection-depth info */
class ibuf
{
friend class ibuf_iter;
    dqueue *list;              /* array of intersections */
    int     len,                /* length of list */
    max;			/* amount of mem alloced for list */
    dqueue minq;		/* closest intersection */
    bool    sorted;             /* sorted? */

    void sort();
    void init()
	{
	    len = 0;
	    minq.obj = minq.obj2 = 0;
	    minq.t = INFTY;
	    sorted = false;
	}
    void remax(int);
public:
    void print() const ;
    ibuf();

    void add_to(int, dqueue const &);
    void add_to(int, flt, object *, bool, object * = 0);
    flt     mint()
	{
	    return minq.t;
	}
    dqueue & min()
	{
	    return minq;
	}
    bool    empty()
	{
	    return !len;
	}
    ~ibuf();
};

class ibuf_iter {
    ibuf *q;
    dqueue *end;
    dqueue *idxp; /* temporary variable */
public:
    ibuf_iter(ibuf &r) {
	q = &r;
	idxp = q->list;
	end = q->list + q->len;

	if (!q->sorted)
	    q->sort();
    }
    operator dqueue * () {
	return idxp;
    }
    dqueue *operator->() {
	return idxp;
    }
    operator bool() {
	return (idxp<end );
    }
    dqueue *next()
	{
	    return (idxp < end) ? (++idxp) : 0;
	}
};


/* structure for passing intersection-point info */
struct intersection {
    dqueue  q_ent;              /* the minimum intersection */
    ray    *r;                  /* the ray */
    vector  n,                  /* the normal at that point */
            x,                  /* the intersection point in world coordinates (with speed) */
            objectx,            /* intersection in shape coordinates */
            objectn;            /* normal in lcs */

};
#endif
