/*
   lights.cc --  routines for lights


   (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */


#include "object.hh"
#include "light.hh"
#include "lightbuf.hh"
#include "proto.hh"
#include "extern.hh"
#include "ibuf.hh"
#include "color.hh"
#include "stats.hh"

int lightbufres = 5;
bool    illumination_data::all_intersections(ibuf *, ray * , int, bool &) const
{
    return false;
}

shape_stats *light_data::stat = shape_stats::regist("light"),

*fog_data::stat =  shape_stats::regist("fog");

/****************************************************************
  LIGHTLISTS
 ****************************************************************/

#include "vray.hh"
svec<object *> lightlist;

light_iter::light_iter() {
    j = 0;
}

bool light_iter::ok() {
    return j < lightlist.sz();
}

void
light_iter::next() {
    j++;
}

object *light_iter::operator->() {
    return lightlist[j] ;
}
light_data *light_iter::data() {
    return(light_data*) lightlist[j]->data ;
}
int
light_list_len()
{
    return lightlist.sz();
}

/****************************************************************
  ILLUM
 ****************************************************************/
illumination_data::illumination_data()
{
    type = LIGHT;
    org.set(0);
    dir.set(1, 0, 0);
    tightness = 0;
    lcolor.set(1);
    attenuation_exp = 0.0;
    radius = 0.0;
}

void
illumination_data::scale(const vector s)
{  
    org.scale(s);

    vector  inv;
    inv.x = 1 / s.x;
    inv.y = 1 / s.y;
    inv.z = 1 / s.z;
    dir.scale(inv);
    dir /= dir.length();
}

void
illumination_data::rotate(const matrix & rotmat)
{
    org.rotate(rotmat);
    dir.rotate(rotmat);
}

void
illumination_data::translate(const vector t)
{
    org += t;
}

/****************************************************************/

void
illumination_data::print() const
{
#ifdef DEBUG
    lcolor.print();
    org.print("origin");
    dir.print("direction");
    dprintf("attenuation %f, radius  %f\n",
            attenuation_exp, radius);
    dprintf("tightness %f ", tightness);
#endif          /* DEBUG */
}


void
light_data::print_list()
{
    #ifdef DEBUG
    for (light_iter f; f.ok(); f.next())
	f->print();
#endif
}

/* add light to list of lights */
void
illumination_data::bound(matrix *, bounding_box *)
{
    
}

illumination_data::~illumination_data() { }

int
light_data::count()
{
    return lightlist.sz();
}
light_data::~light_data()
{
    delete buf;
}

light_data::light_data() 
{
    buf = 0;
    shaded  = shad_test = 0;
}

void
light_data::precompute()
{
    stat->howmuch++;
    lightlist.add(daddy);
    buf = new lightbuf (org);

    prims++;
}

PUBLIC object *
        get_new_light_object()
{
    object *o = new object;
    illumination_data *l = new light_data;
    o->data = l;
    o->flags |= O_INVIS;
    l->setdaddy(o);
    return o;
}

color
light_data::get_color(vector a) const
{
    if (tightness)
    {
        vector  d = a - org;
        d /= d.length();

        flt dot = d * dir;
	/* this can happen, if the lighted point is not in the direction of the spotlight */
        if (dot < 0)  
            dot = 0;

        flt intensity = pow(dot, tightness);
        return intensity * lcolor;
    } else
        return lcolor;
}

bool
light_data::test(ray *r) {
    shad_test++;
    bool b= buf->test(r);
    if (b)
	shaded++;
    return b; 
}

flt light_data::hitratio() {
    if (shad_test)
      return flt(shaded)/flt(shad_test);
    else
	return 0.0;
}
/****************************************************************/

svec<object *> foglist;

void
fog_data::print_list()
{
#ifdef DEBUG
    for (int i =0; i < foglist.sz(); i++)
	foglist[i]->print();
#endif

}

void
fog_data::precompute()
{
    stat->howmuch++;
    foglist.add(daddy);
    prims++;
}

PUBLIC object *
        get_new_fog_object()
{
    object *o = new object;
    illumination_data *l = new fog_data;
    o->data = l;
    o->flags |= O_INVIS;
    o->data->setdaddy(o);
    return o;
}

/* add fog to ray */
color
fog_data::get_color(ray * r, flt endpoint)
{
    Color   col(0);

    /* Walk the light list. */
    for (int i = 0; i < foglist.sz(); i++)
    {
	object *op = foglist[i];
	fog_data *        l = (fog_data *) op->data;
        vector lorg = l->org;
	lorg += r->now * op->getspeed();

        vector pminl = r->pos - lorg;
        flt t0 = -pminl * r->dir;
        flt dirfact = 1;
        vector closepoint = t0 * r->dir + pminl;

        if (l->tightness) {
            vector closedir = closepoint / closepoint.length();
            dirfact = closedir * l->dir;
            if (dirfact < EPSILON)
            {
                continue;
            }
            dirfact = pow(dirfact, l->tightness);
        }

        flt d0sq = closepoint * closepoint;
        flt d0 = sqrt(d0sq);
        if (ISZERO(d0))
            continue;

    flt intensity = (atan((endpoint - t0) / d0) - atan(-t0 / d0)) /
	((d0 + l->radius) * pow((1 + d0), l->attenuation_exp));

        // don't use light_data::get_color, since directional effects are different
        col += (intensity * dirfact) * l->lcolor;
    }
    return col;
}

