/*
  objects.c --  object manipulations


  (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>
  It is based on "Ray" by George Kyriazis <kyriazis@turing.cs.rpi.edu>
  Rayce is released under the terms of the GNU General Public License,
  it has no warranty what so ever. For more details: see the file "COPYING"
*/


#include "object.hh"
#include "shapes.hh"
#include "ibuf.hh"
#include "proto.hh"
#include "extern.hh"
#include "texture.hh"

PRIVATE void print_flags(int flags);

vector
object::getspeed()
{
    return  (speed)  ?  (*speed) :  vector(0);    
}

inline bool
object::intersect_bounds(ibuf *globalq, ray *r, int traceflags, bool &isinside) const
{
    /* check for bounding shapes, walk them all */
    for (object_iter op(bounds); !(traceflags & NO_CHKBOUNDS) && op.ok(); op.next())
    {

        bool    inside_bound;
        ibuf   q;

        /* Check bounding shape: */
        boundstat->test++;

        /* don't translate bounds. */
        if (!op->all_intersections(&q, r, CHKINSIDE, inside_bound))
        {
            /* no intersection at all */
            if (traceflags & CHKINSIDE)
                isinside = inside(r->pos);

            return false;
        }
        if (inside_bound)
        {

            continue;
        }
        /* found one! */
        if (!(traceflags & CHKALL) && q.mint() >= globalq->mint())
        {                       /* close enough? */
            if (traceflags & CHKINSIDE)
                isinside = inside(r->pos);

            return false;
        }

        boundstat->hit++;       /* And another bound has been succesful */
    }
    return true;
}


inline bool
object::intersect_clips(ibuf *globalq, ray *localray, ray *r, int traceflags, bool &isinside) const
{
    /* do clippingshape */ 
    ibuf q;
    bool gotcha = data->all_intersections(&q, localray, traceflags | CHKALL, isinside);
    if (!gotcha)
	return false;

    bool    hit = false;
    for (ibuf_iter qp(q);
	 qp && (traceflags & CHKALL || !hit);
	 qp.next())
        {
            bool    inside_clip;
            /* find intersection point */
            vector  x = r->point(qp->t);
            clipstat->test++;

            /* filter out the intersections which are inside of the clipping shape. */
            inside_clip = true;
            for (object_iter op(clip); op.ok() && inside_clip; op.next())
		{
		    if (!op->inside(x))
			inside_clip = false;
		}

            if (inside_clip)
		{
		    globalq->add_to(traceflags, *(dqueue*) qp);
		    clipstat->hit++;
		    hit = true;
		}
        }

    if (traceflags & CHKINSIDE)
	isinside = inside(r->pos);
    return hit;
}

/*
 * intersect a ray with an object.
 */
bool
object::all_intersections(ibuf * globalq, ray * r, int traceflags, bool &isinside) const
{
    stat->test++;

    /* mailboxes */
    if ((traceflags & USEMAIL) && last_ray_id == r->rayid)
        return false;
    else
        ((object *) this)->last_ray_id = r->rayid;

    /* the o_no_xxx flags */
    if (r->type & flags)
        return false;

    /* needed for correct refraction */
    if (flags & O_TRANSPARENT)
    {
        traceflags |= CHKINSIDE;

        /* will be needed for attenuation of shadowrays */
        if (r->type == R_SHADOW)
            traceflags |= CHKALL;
    }
    ray     transray = *r;

    /* use speed in the ray. */
    if (speed) {
        transray.pos += -r->now * *speed;

        /* tricky: throw away pointer to original input ray */
        r = &transray;
    }
    
    if (!intersect_bounds(globalq, r, traceflags, isinside))
	return false;
    
    traceflags &= ~NO_CHKBOUNDS;

    ray     localray( *r);
    transformr(localray);

    bool    hit;
    if (!clip.empty())
	hit = intersect_clips(globalq, &localray, r, traceflags, isinside);    
    else 
        /* intersect the ray with object's shape */
        hit = data->all_intersections(globalq, &localray, traceflags, isinside);

    if (hit)
        stat->hit++;
    return hit;
}

/* apply speed to one object */
void
object::add_speed(vector s)
{
    if (data->add_speed != (void *) shape_data::add_speed && !data->refcount)
        data->add_speed(s);
    else {
        if (!speed)
            speed = new vector(0);
        *speed += s;
    }
}

void
object::append_to(object * toadd)
{
    assert(!toadd->next);
    object *last = 0;
    for (object * o = this; o; o = o->next)
	last = o;
    
    last->next = toadd;
}

/* add clipping shape from boundingshapes. */
void
object::add_bounds_to_clip()
{
    if (!bounds.empty()) {                           /* this object has a bound. */
        for (object_iter op(bounds); op.ok(); op.next()) {
	    object *dst = new object(*op);
            clip.append_to(dst);
	    dst->daddy = this;
        }
    }
}

object::object (object & src) : tranp(src.tranp), bounds(src.bounds), clip(src.clip)
{    
    //init
    next = 0;
    last_ray_id = -1;

    /* copy info */
    flags = src.flags;

    /* copy pointers: texture */
    text  = (src.text)  ? src.text->clone()      :0;
    bbox  = (src.bbox)  ? new bounding_box(*src.bbox) :0;
    speed = (src.speed) ? new vector(*src.speed)      :0;

    /* copy the shape */
    if (src.data->refcount) {
	data = src.data;
	assert(!data->setdaddy);
	data->refcount ++;
    } else {
	data = src.data->clone();	
	data->setdaddy(this);
    }
    data->type = src.data->type; //overbodig?
}

void
object::bound()
{
    /* allocate a bounding box, if necessary. */
    if( (flags & O_INFINITE) || (flags & O_INVIS) || bbox )
	return;
    else
        bbox = new bounding_box;
    if (data->refcount) {
	data->bound(0,bbox);
    } else
	data->bound(tranp.trans, bbox);

    /* take intersection of Bboxes */
    bounds.bound();
    for (object_iter op(bounds); op.ok(); op.next())
    {
        if (!(op->flags & O_INFINITE)) {
	    bbox->intersect(*op->bbox);
        }
    }

    /* take intersection of clips */
    clip.bound();
    for (object_iter op(clip); op.ok(); op.next())
    {
        if (!(op->flags & O_INFINITE)) {
	    bbox->intersect(*op->bbox);
	}
    }

    // into shape?
    /* incorporate speed into the bounds */
    if (speed)
	bbox->add_speed(*speed);


    bbox->scale(BOUNDFUDGE + 1);
}

void
object::precompute()
{
    stat->howmuch++;

    /* ... Do It ... */
    data->precompute();

    /* can I see clear now ? (check transparency) */
    // whugh.. using structures outside "this"
#if 0
    object *dp;
    for (dp = this; dp && !dp->text; dp = dp->daddy)
        ;

    if (!dp) {
           /* a clipping or bounding shape. */
#ifdef DEBUG
        if (debug_options & DEBUGSELFCHECK) {
            short_print();
            assert(false);
        }
#endif
    } else if (dp->text->finish.refr_color.length() > EPSILON) {
	flags |= O_TRANSPARENT;
	if (!(flags & O_CLOSED) && data->type != COMPOSITE)
	    warning("Transparent objects should be closed");
    }
#endif // 0

    /* flip inside and outside */
    if ((flags & O_INVERTED)) {
        int     fi,
                fo;
        fo = flags & O_FINITE_OUTSIDE;
        fi = flags & O_FINITE_INSIDE;

        if (fi)
            flags |= O_FINITE_OUTSIDE;
        else
            flags &= ~O_FINITE_OUTSIDE;

        if (fo)
            flags |= O_FINITE_INSIDE;
        else
            flags &= ~O_FINITE_INSIDE;
    }

    /* stats */
    if (!bounds.empty()) {

	/* take intersection of Bboxes */
	bounds.precompute(this);
	for (object_iter op(bounds); op.ok(); op.next())
	    if (!(op->flags & O_INFINITE))
		flags &= ~O_INFINITE;

        boundstat->howmuch++;
    }

    if (!clip.empty()) {
	clip.precompute(this);
	for (object_iter op(clip); op.ok(); op.next())
	    if (!(op->flags & O_INFINITE))
		flags &= ~O_INFINITE;
        clipstat->howmuch++;
    }
}

object::object()
{
    speed = 0;
    bbox = 0;
    text = 0;
    data = 0;
    next = 0;
    daddy = 0;
    flags = 0;
    last_ray_id = -1;
}

/* get normal to object. Handles inverse transforms */
vector
object::normal(vector loc) const
{
    loc = transformx(loc);
    vector      n = data->normal(loc);
    n = transformn(n);

    n.normalise();
    return n;
}

/* is "loc" inside this object? Handles inverse transforms */
bool
object::inside(vector loc) const
{
    bool        inside_data = data->inside( transformx(loc));

    if (flags & O_INVERTED)
        return !inside_data;
    else
        return inside_data;
}

void
object::rotate(const matrix & rotmat)
{
    if (text)
        text->rotate(rotmat);

    if (data->rotate != (void*)shape_data::rotate && !tranp.trans && !data->refcount)
	data->rotate(rotmat);
    else
	tranp.rotate(rotmat);

    if (speed)
        speed->rotate(rotmat);

    bounds.rotate(rotmat);
    clip.rotate(rotmat);
}

void
object::translate(vector t)
{
    if (text)
        text->translate(t);

    if (tranp.inv_trans ||  data->translate == (void*) shape_data::translate)
	tranp.translate(t);
    else
	data->translate(t);

    clip.translate(t);
    bounds.translate(t);
}

/* and scale it */
void
object::scale( vector s)
{
    if (tranp.inv_trans || data->scale == (void*)shape_data::scale ||
	( (flags & O_ONLY_UNIFORM_SCALE) && (abs(s.x) != abs(s.y) || abs(s.y)!= abs(s.z) ) )
	||data->refcount)
	tranp.scale(s);
    else
	data->scale(s);

    if (text)
        text->scale(s);

    if (speed)
        speed->scale(s);

    bounds.scale(s);
    clip.scale(s);
}


object::~object()
{
    delete text;
    if (data->refcount)
	if (!data->refcount--)
	    delete data;
    delete bbox;
}

void
object::print() const
{
#ifdef DEBUG
    dprintf("object {\n");
    print_flags(flags);
    if (!bounds.empty()) {
        dprintf("BOUNDINGSHAPES:");
        bounds.print();
    }
    if (!clip.empty()) {
        dprintf(" CLIPPINGSHAPES:");
        clip.print();
    }

    if (bbox)
    {
        dprintf("BBOX:\n");
	bbox->print();
    }

    tranp.print();
    dprintf("shape type %d", data->type);
    if (data->type != NOSHAPE) {
        dprintf(" %s (ref %d) {\n", (data->getstatp())->name, data->refcount);
        data->print();
	dprintf("}\n");
    }

    if (text)
        text->print();
    if (speed)
        speed->print("speed");
    dprintf("}\n");
#endif          /* DEBUG */
}

void
object::short_print() const
{
#ifdef DEBUG
    dprintf("[");
    if (bbox)
    {
        dprintf("BBOX:\n");
        bbox->print();
    }
    dprintf("shape type %d, %s]\n", data->type, (data->getstatp())->name);
#endif          /* DEBUG */
}

/****************************************************************/

/* print out object flags */
PRIVATE void
print_flags(int flags)
{
#ifdef DEBUG
    dprintf("object flags: ");
    if (flags & O_NO_SHADOW)
        dprintf("noshad ");
    if (flags & O_NO_REFLECT)
        dprintf("norefl");
    if (flags & O_NO_REFRACT)
        dprintf("norefr ");
    if (flags & O_NO_EYE)
        dprintf("noeye ");
    if (flags & O_TRANSPARENT)
        dprintf("transp ");
    if (flags & O_INFINITE)
        dprintf("infin ");
    if (flags & O_INVIS)
        dprintf("invis ");
    if (flags & O_CLOSED)
        dprintf("closed ");
    if (flags & O_FINITE_INSIDE)
        dprintf("FinIns ");
    if (flags & O_FINITE_OUTSIDE)
        dprintf("FinOut ");
    if (flags & O_INVERTED)
        dprintf("Inv ");
    if (flags & O_ONLY_UNIFORM_SCALE)
	dprintf("unfscale ");
    dprintf("\n");
#endif          /* DEBUG */
}

shape_stats *object::clipstat = shape_stats::regist("clip"),
       *object::boundstat = shape_stats::regist("bound"),
       *object::stat = shape_stats::regist("object");
