/*
   Shade.c -- the shading function

   (c) 1994, 1995 by Shawn McHorse <Zaphod@fcircus.sat.tx.us> and Han-Wen
   Nienhuys <hanwen@stack.urc.tue.nl>

   (Shawn McHorse <smchorse@ringer.cs.utsa.edu> (?))

   Based on code by George Kyriazis

   this is part of Rayce, a raytracer based on "Ray" by George Kyriazis
   <kyriazis@turing.cs.rpi.edu> Rayce is released under the terms of the
   GNU General Public License, it has no warranty what so ever. For more
   details: see the file "COPYING"
 */

#include "light.hh"
#include "proto.hh"
#include "ibuf.hh"
#include "texture.hh"
#include	"extern.hh"
#include 	"object.hh"
#include 	"light.hh"
#include 	"grid.hh"
#include "vray.hh"
#include "stats.hh"

/****************************************************************
  LIGHTSOURCE LIGHTING
 ****************************************************************/
/*

   contribution from each lightsource to local shading this is
   computed *before* the actual shadow testing, so we can ignore the
   contribution if it is small compared to other contribs.

   Gregory J Ward, Photorealism in rendering, " " 

*/   
class light_contrib { 	// contribution of ONE lightsource
public:    
    color c;
    flt clen,ldist;
    ray r;		// shadowray
    light_data *l;

    // can't use real constructor due to array
    light_contrib():r(R_SHADOW, 0.0) {	
	c.set(0);
	clen =0.0;
	l = 0; 
    }
    void init(light_data *o, intersection *i,local_texture *tp, Grid_tree *g);
    color get_color() { return c; }
};

// calculate contribution
void 
light_contrib::init(light_data *o, intersection *i, local_texture *tp, Grid_tree *g)
{
    l = o;

    vector  	lorg = l->org;
    r.now = i->r->now;
    lorg +=  r.now * l->daddy->getspeed();

    /* Construct a normalized vector from object to light source. */
    r.dir = lorg - i->x;
    ldist = r.dir.length();
    
    if (ldist < tolerance)
	return;
    
    r.mint = ldist;
    r.maxt = ldist+10*tolerance;		  // who the fuck is 10
    
    r.dir /= ldist;
    r.pos = i->x;
    
    flt	filter_factor;

    sample_ray(&r, &filter_factor, l->radius / ldist, g);

    /* Check if the light source is behind the object. */
    flt    ldotn = r.dir * i->n;

    if (ldotn < EPSILON) {	
	return;
    }

    /* Diffuse shading. */
    if (tp->diffuse) {
	c += pow(ldotn, tp->brilliance) * tp->diffuse;
    }

    /* Specular shading. */
    if (tp->specular && !(l->daddy->flags & O_NO_REFLECT)) {
	flt invr = (tp->roughness > EPSILON) ? 1/tp->roughness : INFTY;
	
	flt f = tp->distfunc(invr, r.dir, i->r->dir, i->n);
	f *= tp->geofunc(r.dir, i->r->dir, i->n);
	
	c += f * tp->specular;
    }
    c *= l->get_color(i->x);
    if (l->attenuation_exp > EPSILON)
	c *= pow(ldist, - l->attenuation_exp);
    clen = c.length();
}

// influence of all lightsources.
struct light_contribs {
    int llen;
    light_contrib *clist;
    flt total;
    light_contribs(intersection * i, local_texture * tp, Grid_tree * g);
    color get_color() const;
    ~light_contribs() { delete[] clist ; }
};

flt light_eps=0.01;

// should enable darklights.
static shape_stats *fudge = shape_stats::regist("light fudge", -1);
color
light_contribs::get_color() const
{
    flt current_illum=0.0;
    color retcol(0);


    // test each lightsource, from large contributions to small ones
    // try to get out early in order save time
    int j =0;
    light_contrib *lcp = clist;
    for (; j < llen && lcp->clen > 0 ; j++,lcp++) {
	bool b = lcp->l->test(&lcp->r);
	fudge->test++;
	if (!b) {
	    current_illum += lcp->clen;
	    retcol += lcp->c;

	    if ((total-current_illum) / total  < light_eps)
		break;
	}
    }
    
    // to avoid aliasing add some color from other ls's
    for (; j < llen; j++,lcp++) {
	if (lcp->l->test && lcp->clen > 0) {
	    fudge->hit++;
	    fudge->test++;
	    retcol += lcp->c * lcp->l->hitratio();
	}
    }
    return retcol;
}

// compare function for qsort
static int lc_comp(void const *p1, void const *p2)
{
    return (int) SGN( ( (light_contrib const *) p2)->clen -((light_contrib const *) p1)->clen );
}

light_contribs::light_contribs(intersection * i, local_texture * tp, Grid_tree * g)
{
    llen = light_list_len();
    clist = new light_contrib [llen];

    int j =0;
    total = 0;
    for (light_iter op; op.ok(); op.next(),j++) {
 	clist[j].init(op.data(),i,tp,g);
	total += clist[j].clen;
    }
    qsort((void *)clist, llen, sizeof(light_contrib), lc_comp);
}

PRIVATE color
local_shading(intersection * i, local_texture * tp, Grid_tree * g)
{
    ray    *r = i->r;
    /* Start off with ambient light. */
    color       retcol = *the_scene.ambient_light * tp->ambient;

    /* the normal should be looking at us */
    
    if (i->n * r->dir > 0)
        i->n *= -1;           /* flip it */

    light_contribs lc(i,tp,g);
    retcol += lc.get_color();
    return retcol;
}

/****************************************************************/
  
/* global shading model
   PARAMS

   i:	ray/obj intersection
   g:	recursion info
   imp:	part to pixelcolor.

   RETURN:
   local_shading + reflected + refracted

 */
color
Scene::shade(intersection * i, Grid_tree * g,
	      flt imp)
{                               /* imp is for the depth cutoff */    
    flt     importance,
            filter_factor;

    /* Find the first object in the hierarchy with a texture. */
    object *op = i->q_ent.obj;
    
    for (; !op->text; op = op->daddy);

    assert(op);

    /* Find texture at intersection point. */
    local_texture t=  op->text->get_local_texture(i->x, i->n);
    color col = local_shading(i, &t, g);

    /* Don't bother with reflected and refracted rays if we're already as deep as we can go. */
    if (g->recursion >= max_trace_level)
        return col;

    importance = imp * t.refr_color.length();
    flt c1(0.0);
    if (importance > depth_cutoff)
    {
        /* ***************REFRACTION************** */

        ior_depth++;
        if (i->q_ent.entering)
            global_ior[ior_depth] = t.index;
        else // exiting material : pop ior. 
            global_ior[ior_depth] = global_ior[ior_depth - 2];

        /* Calculate refracted vector. */
        c1 = -i->r->dir * i->n;
        flt nu = global_ior[ior_depth] / global_ior[ior_depth - 1];
        flt discr = nu * nu - 1 + c1 * c1;

        /* Total Internal Reflection */
        if (discr < -EPSILON)
        {

            /* Calculate reflected vector. */
            vector  refl = 2 * c1 * i->n;
            refl = i->r->dir + refl;

            /* Set up reflected ray. */
            ior_depth--;
            ray     reflected(R_REFLECT, i->r->now);
            reflected.dir = refl;
            reflected.pos = i->x;
//	    reflected.mint = INFTY;	 // HUH!?


            /* we're reflecting and refracting at the same time, so take means */
            flt     mean = (t.refl_diffuse + t.refr_diffuse) / 2.0;
            sample_ray(&reflected, &filter_factor, mean, g);

            color tracecol =  trace(reflected, g->reflect, importance + imp * t.refl_color.length());
            color   tircol = t.refr_color + t.refl_color;

	    /* 0.5: reflecting AND refracting */
            col += tircol*(0.5 * filter_factor * tracecol);
            return col;
        }
        vector  refr=         (discr < EPSILON) ? c1 * i->n : (c1 - sqrt(discr)) * i->n;

        refr = i->r->dir + refr;
        refr = 1 / nu * refr;

        /* Set up the refracted ray. */
        ray     refracted(R_REFRACT, i->r->now);
        refracted.dir = refr;
        refracted.pos = i->x;

        sample_ray(&refracted, &filter_factor, t.refr_diffuse, g);
        color tracecol = trace(refracted, g->refract, importance);

        col = col + t.refr_color * filter_factor*tracecol;
        ior_depth--;
    }
    importance = imp * t.refl_color.length();

    if (importance > depth_cutoff)
    {
        /* ***************REFLECTION************** */

        /* Calculate reflected vector. */
        if (c1 == 0.0)
            c1 = -i->r->dir * i->n;
        vector  refl = 2 * c1 * i->n;
        refl = i->r->dir + refl;

        /* Set up the reflected ray. */
        ray     reflected(R_REFLECT, i->r->now);
        reflected.dir = refl;
        reflected.pos = i->x;


        sample_ray(&reflected, &filter_factor, t.refl_diffuse, g);
        color tracecol = filter_factor * trace(reflected, g->reflect, importance);

        col = col + t.refl_color * tracecol;
    }
    return col;
}

