/*
  vector.h -- inline vector operations, definitions

  (c) 1993, 1994, 1995 Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>
  It is based on "Ray" by George Kyriazis <kyriazis@turing.cs.rpi.edu>
  Rayce is released under the terms of the GNU General Public License,
  it has no warranty what so ever. For more details: see the file "COPYING"

 */
#ifndef VECTOR_H
#define VECTOR_H
#include "ray.hh"


/// guess what, without constructor 
struct nc_vector
{
    flt     x;
    flt y;
    flt z;

    /// index as if array.
    flt & operator[] (int i) const {
	assert(0 <= i < 3);
	return ((flt *) this ) [ i ]; // megaugly
    }

    /// find largest component.
    axes dominant() const;

    /// set each component
    void     set(const flt nx, const flt ny, const flt nz)
    {
        x = nx;
        y = ny;
        z = nz;
    }
    /// set all components
    void set(const flt a)
    {
        x = y = z = a;
    }

    void print(const char *s) const;
    void print() const
    {
        print("");
    } 
    /// add an another vector
    nc_vector & operator += (const nc_vector & a)
    {
        x += a.x;
        y += a.y;
        z += a.z;
        return *this;
    }
    /// scalar divide
    nc_vector & operator /= (flt a)
    {
        x /= a;
        y /= a;
        z /= a;
        return *this;
    }

    nc_vector & operator -= (const nc_vector & a)
    {
        x -= a.x;
        y -= a.y;
        z -= a.z;
        return *this;
    }
    /// scalar multiply
    void operator *= (flt  a)
    {
        x *= a;
        y *= a;
        z *= a;

    }
    /// 1-norm.
    flt     quicklength() const
    {
        flt     s;
                s = (x < 0) ? -x : x;
                s += (y < 0) ? -y : y;
                s += (z < 0) ? -z : z;
                return s;
    }
    /// euclidean length
    flt     length() const
    {
        return sqrt(x * x + y * y + z * z);
    }

    /// nullvector?
    operator bool() const 
    {
        return (quicklength() > EPSILON);
    }

    /// set to length 1
    void normalise() {
	*this /= length();
    }
};

inline nc_vector operator - (const nc_vector & a)
{
    nc_vector r;
    r.x = -a.x;
    r.y = -a.y;
    r.z = -a.z;
    return r;
}


inline nc_vector operator + (const nc_vector & v1, const nc_vector & v2)
{
    nc_vector res = v1;
    res += v2;
    return res;
}

/* crossproduct */
inline nc_vector operator ^ (const nc_vector & v1, const nc_vector & v2)
{
    nc_vector res;
    res.set(v1.y * v2.z - v1.z * v2.y
            ,v1.z * v2.x - v2.z * v1.x,
            v1.x * v2.y - v2.x * v1.y);
    return res;
}

inline nc_vector operator - (const nc_vector & v1, const nc_vector & v2)
{
    nc_vector res = v1;
    res -= v2;
    return res;
}

inline int operator == (const nc_vector & v1, const nc_vector & v2)
{
    return  v1 - v2;
}

inline flt operator *(const nc_vector & v1, const nc_vector & v2)
{
    return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

inline nc_vector operator *(const flt & a, const nc_vector & v2)
{
    nc_vector res = v2;
    res *= a;
    return res;
}

inline nc_vector operator *(const nc_vector & v2, const flt & a)
{
    return a * v2;
}

inline nc_vector operator / (const nc_vector & v2, const flt & a)
{
    return v2 * (1 / a);
}

/// vector with constructor
struct vector : nc_vector
{
    /// auto typecast from nc_vector
    vector(const nc_vector & v) {
        x = v.x;
        y = v.y;
        z = v.z;
    }
    vector() { }
    vector(const flt nx, const flt ny, const flt nz)
    {
        x = nx;
        y = ny;
        z = nz;
    }
    vector(const flt a)
    {
        x = y = z = a;
    }
    void     rotate(const matrix & m);
    void     scale(const vector s)
    {
        x *= s.x;
        y *= s.y;
        z *= s.z;
    }
    
};

struct direction : vector
{
    // auto typecast from nc_vector
    direction(const vector & v)
    {
        x = v.x;
        y = v.y;
        z = v.z;
    }
    direction()
    {

    }
};

extern vector bggradient;       /* gradient in which direction? */


#endif
